#pragma once

/* Generated with cbindgen:0.9.0 */

/* DO NOT MODIFY MANUALLY! Changes will be discarded during build. */

#include <cstdarg>
#include <cstdint>
#include <cstdlib>
#include <new>
#include <memory>

/// Enum to represent log levels in the application
enum class LogLevel {
  Error = 0,
  Warn = 1,
  Info = 2,
  Debug = 3,
  Trace = 4,
};

/// Enum used to specify scaling quality for scaling up graphics
enum class ScalingQuality {
  /// Use linear interpolation
  LINEAR,
  /// Use nearest neighbor interpolation
  NEAR_PERFECT,
  /// Scale up to the nearest multiple of 640x480 and use nearest neighbor interpolation
  PERFECT,
};

/// Enum for the vanilla game version that is used to run the game
enum class VanillaVersion {
  DUTCH,
  ENGLISH,
  FRENCH,
  GERMAN,
  ITALIAN,
  POLISH,
  RUSSIAN,
  RUSSIAN_GOLD,
};

/// Struct that is used to store the engines configuration parameters
struct EngineOptions;

/// Thread safe library database.
struct LibraryDB;

/// Library file.
/// Provides access to the data of a library entry through Read and Seek.
struct LibraryFile;

/// A wrapper around `Vec<CString>` for C.
struct VecCString;

extern "C" {

/// Deletes a CString.
/// The caller is no longer responsible for the memory.
void CString_destroy(char *s);

/// Clears `EngineOptions.mods`.
void EngineOptions_clearMods(EngineOptions *ptr);

/// Creates `EngineOptions` with the provided command line arguments.
/// Loads values from `(stracciatella_home)/ja2.json`, creating it if it does not exist.
/// The caller is responsible for the returned memory.
EngineOptions *EngineOptions_create(const char *const *args, size_t length);

/// Deletes `EngineOptions`.
void EngineOptions_destroy(EngineOptions *ptr);

/// Gets `EngineOptions.brightness`.
float EngineOptions_getBrightness(const EngineOptions *ptr);

/// Gets the target index of `EngineOptions.mods`.
/// The caller is responsible for the returned memory.
char *EngineOptions_getMod(const EngineOptions *ptr, uint32_t index);

/// Gets the length of `EngineOptions.mods`.
uint32_t EngineOptions_getModsLength(const EngineOptions *ptr);

/// Gets the width of `EngineOptions.resolution`.
uint16_t EngineOptions_getResolutionX(const EngineOptions *ptr);

/// Gets the height of `EngineOptions.resolution`.
uint16_t EngineOptions_getResolutionY(const EngineOptions *ptr);

/// Gets `EngineOptions.resource_version`.
VanillaVersion EngineOptions_getResourceVersion(const EngineOptions *ptr);

/// Gets `EngineOptions.scaling_quality`.
ScalingQuality EngineOptions_getScalingQuality(const EngineOptions *ptr);

/// Gets the `EngineOptions.stracciatella_home` path.
/// The caller is responsible for the returned memory.
char *EngineOptions_getStracciatellaHome(const EngineOptions *ptr);

/// Gets the `EngineOptions.vanilla_game_dir` path.
/// The caller is responsible for the returned memory.
char *EngineOptions_getVanillaGameDir(const EngineOptions *ptr);

/// Adds a mod to `EngineOptions.mods`.
void EngineOptions_pushMod(EngineOptions *ptr, const char *name);

/// Sets `EngineOptions.brightness`.
void EngineOptions_setBrightness(EngineOptions *ptr, float brightness);

/// Sets `EngineOptions.resolution`.
void EngineOptions_setResolution(EngineOptions *ptr, uint16_t x, uint16_t y);

/// Sets `EngineOptions.resource_version`.
void EngineOptions_setResourceVersion(EngineOptions *ptr, VanillaVersion res);

/// Sets `EngineOptions.scaling_quality`.
void EngineOptions_setScalingQuality(EngineOptions *ptr, ScalingQuality scaling_quality);

/// Sets `EngineOptions.start_in_fullscreen`.
void EngineOptions_setStartInFullscreen(EngineOptions *ptr, bool val);

/// Sets `EngineOptions.start_without_sound`.
void EngineOptions_setStartWithoutSound(EngineOptions *ptr, bool val);

/// Sets the `EngineOptions.vanilla_game_dir` path.
void EngineOptions_setVanillaGameDir(EngineOptions *ptr, const char *game_dir_ptr);

/// Gets `EngineOptions.run_editor`.
bool EngineOptions_shouldRunEditor(const EngineOptions *ptr);

/// Gets `EngineOptions.run_unittests`.
bool EngineOptions_shouldRunUnittests(const EngineOptions *ptr);

/// Gets `EngineOptions.show_help`.
bool EngineOptions_shouldShowHelp(const EngineOptions *ptr);

/// Gets `EngineOptions.start_in_debug_mode`.
bool EngineOptions_shouldStartInDebugMode(const EngineOptions *ptr);

/// Gets `EngineOptions.start_in_fullscreen`.
bool EngineOptions_shouldStartInFullscreen(const EngineOptions *ptr);

/// Gets `EngineOptions.start_in_window`.
bool EngineOptions_shouldStartInWindow(const EngineOptions *ptr);

/// Gets `EngineOptions.start_without_sound`.
bool EngineOptions_shouldStartWithoutSound(const EngineOptions *ptr);

/// Writes `EngineOptions` to `(stracciatella_home)/ja2.json`.
/// Returns true on success, false otherwise.
bool EngineOptions_write(EngineOptions *ptr);

/// Constructor.
/// The caller is responsible for the memory of the library database.
LibraryDB *LibraryDB_create();

/// Destructor.
/// The caller is no longer responsible for the memory of the library database.
void LibraryDB_destroy(LibraryDB *ldb);

/// Adds a library to the end of the library database.
/// Returns true is successful, false otherwise.
/// Sets the rust error.
bool LibraryDB_push(LibraryDB *ldb, const char *data_dir, const char *library);

/// Closes a library database file.
/// The caller is no longer responsible for the library file memory.
void LibraryFile_close(LibraryFile *file);

/// Gets the current position in a library database file.
uint64_t LibraryFile_getPosition(LibraryFile *file);

/// Gets the size of a library database file.
uint64_t LibraryFile_getSize(LibraryFile *file);

/// Opens a library database file for reading.
/// Returns the file on success and null on error.
/// The caller is responsible for the library file memory.
/// Sets the rust error.
LibraryFile *LibraryFile_open(LibraryDB *ldb, const char *path);

/// Reads from a library database file.
/// Sets the rust error.
bool LibraryFile_read(LibraryFile *file, uint8_t *buffer, size_t buffer_length);

/// Seeks a library database file.
/// `from` expects FileSeekMode values (see src/sgp/SGPFile.h)
/// Sets the rust error.
bool LibraryFile_seek(LibraryFile *file, int64_t distance, int from);

/// Initializes the logger
void Logger_initialize(const char *log_file);

/// Log with custom metadata
void Logger_log(LogLevel level, const char *message, const char *target);

/// Set log level
void Logger_setLevel(LogLevel level);

/// Gets the string representation of the `ScalingQuality` value.
/// The caller is responsible for the returned memory.
char *ScalingQuality_toString(ScalingQuality quality);

/// Gets the string represntation of the `VanillaVersion` value.
/// The caller is responsible for the returned memory.
char *VanillaVersion_toString(VanillaVersion version);

/// Deletes the vector.
void VecCString_destroy(VecCString *vec);

/// Returns the string at the target index.
/// The caller is responsible for the returned memory.
char *VecCString_get(VecCString *vec, size_t index);

/// Returns the vector length.
size_t VecCString_length(VecCString *vec);

/// Returns true if it was able to find path relative to base.
/// Makes caseless searches one component at a time.
bool checkIfRelativePathExists(const char *base, const char *path, bool caseless);

/// Returns a list of available mods.
/// The caller is responsible for the returned memory.
VecCString *findAvailableMods();

/// Converts the launcher executable path to the game executable path.
/// The executable is assumed to be in the same directory as the launcher.
/// The caller is responsible for the returned memory.
char *findJa2Executable(const char *launcher_path_ptr);

/// Finds a path relative to the assets directory.
/// If path is null, it returns the assets directory.
/// If test_exists is true and the path does not exist, it returns null.
/// The caller is responsible for the returned memory.
char *findPathFromAssetsDir(const char *path, bool test_exists);

/// Finds a path relative to the stracciatella home directory.
/// If path is null, it finds the stracciatella home directory.
/// If test_exists is true, it makes sure the path exists.
/// The caller is responsible for the returned memory.
char *findPathFromStracciatellaHome(const char *path, bool test_exists);

/// Returns a copy of the thread local rust error or null.
/// The caller is responsible for the memory.
char *getRustError();

/// Guesses the resource version from the contents of the game directory.
/// Returns a VanillaVersion value if it was sucessful, -1 otherwise.
int guessResourceVersion(const char *gamedir);

} // extern "C"

/// Callable that deletes rust memory.
struct RustDelete
{
	void operator()(char* ptr) const { CString_destroy(ptr); }
	void operator()(EngineOptions* ptr) const { EngineOptions_destroy(ptr); }
	void operator()(LibraryDB* ptr) const { LibraryDB_destroy(ptr); }
	void operator()(LibraryFile* ptr) const { LibraryFile_close(ptr); }
	void operator()(VecCString* ptr) const { VecCString_destroy(ptr); }
};

/// Alias to a std::unique_ptr that deletes with RustDelete.
///
/// By using this alias you are delegating memory management to the C++ compiler.
template<typename T>
using RustPointer = std::unique_ptr<T, RustDelete>;
